/*
 * Synergy KB/PD driver (based on rsdrv sample)
 *
 * (C)Copyright 2008, Yashiro Takeshi
 */

#include <basic.h>
#include <device/kbpd.h>
#include "syndrv.h"


/*
 * ɥ쥹֤ȥå
 */
static ERR setAndCheckTaskSpace(DevReq *devReq)
{
  ERR	err;

  if (devReq->cmd.adcnv == 0 || devReq->datacnt <= 0) {
    return ER_OK;
  }

  /* ɥ쥹֤ */
  err = SetTaskSpace(devReq->taskid);
  if (err != ER_OK) {
    DEBUG_PRINT(("SetTaskSpace err = %d\n", err));
    return err;
  }

  /* ɥ쥹֤Υå */
  if (devReq->cmd.cmd == DC_READ) {
    err = CheckSpaceRW(devReq->memptr, devReq->datacnt);
  }
  else {
    err = CheckSpaceR(devReq->memptr, devReq->datacnt);
  }
  if (err != ER_OK) {
    DEBUG_PRINT(("CheckSpace err = %d\n", err));
    return err;
  }

  return ER_OK;
}


/*
 * ǡȤΥå
 *	return < 0 : 顼
 *	       > 0 : devReq->datacnt;
 */
WERR CheckDataCnt(DevReq *devReq, W datasz)
{
  W reqcnt = devReq->datacnt;

  if (reqcnt < datasz) {
    if (reqcnt != 0) {
      return toERR(EC_PAR, ED_DATACNT);
    }
  }
  return reqcnt;
}


/*
 * ǡɤ߽Ф
 */
static ERR readData(SynergyDriverInfo *inf, 
                    DevRsp *devRsp, DevReq *devReq)
{
  ERR	err;

  /* ѥ᡼å */
  if (devReq->datacnt < 0) {
    err = toERR(EC_PAR, ED_DATACNT);
    goto err_ret;
  }

  /* ɥ쥹֤ȥå */
  err = setAndCheckTaskSpace(devReq);
  if (err < ER_OK) {
    goto err_ret;
  }

  err = toERR(EC_PAR, ED_DATANO);

err_ret:
  DEBUG_PRINT(("readData err = %d\n", err));
  return err;
}


/*
 * ǡ񤭹
 */
static ERR writeData(SynergyDriverInfo *inf, 
                     DevRsp *devRsp, DevReq *devReq)
{
  ERR	err;

  /* ѥ᡼å */
  if (devReq->datacnt < 0) {
    err = toERR(EC_PAR, ED_DATACNT);
    goto err_ret;
  }

  /* ɥ쥹֤ȥå */
  err = setAndCheckTaskSpace(devReq);
  if (err < ER_OK) {
    goto err_ret;
  }

  switch (devReq->datano) {
  case DN_SYNERGY_PDMOVE:
    {
      SynergyDriverPdMove* req = NULL;

      err = CheckDataCnt(devReq, sizeof(SynergyDriverPdMove));
      if (err < 0) {
        goto err_ret;
      }
      else if (err == 0) {
	err = sizeof(SynergyDriverPdMove);
        break;
      }

      req = (SynergyDriverPdMove*)(devReq->memptr);

      inf->evpd.stat.read  = 0;
      inf->evpd.stat.abs   = 1;
      inf->evpd.stat.vst   = (req->enter ? 1 : 0);
      inf->evpd.stat.inv   = 0;
      inf->evpd.stat.nodsp = 0;
      inf->evpd.xpos = req->x;
      if (inf->evpd.xpos < 0) {
        inf->evpd.xpos = 0;
      }
      else if (inf->evpd.xpos >= PDIN_XMAX) {
        inf->evpd.xpos = PDIN_XMAX - 1;
      }
      inf->evpd.ypos = req->y;
      if (inf->evpd.ypos < 0) {
        inf->evpd.ypos = 0;
      }
      else if (inf->evpd.ypos >= PDIN_YMAX) {
        inf->evpd.ypos = PDIN_YMAX - 1;
      }

      /*
       * ܥɥ饤ФǤϡԤʤΤǡKB/PDɥ饤Ф
       * ɬ⤤ͥ٥롼פ˽°ʤФʤʤ
       */
      snd_msg(inf->kpinput, (T_MSG*)&(inf->evpd));

      err = sizeof(SynergyDriverPdMove);
    }
    break;
  case DN_SYNERGY_PDRELMOVE:
    {
      SynergyDriverPdRelMove* req = NULL;

      err = CheckDataCnt(devReq, sizeof(SynergyDriverPdRelMove));
      if (err < 0) {
        goto err_ret;
      }
      else if (err == 0) {
	err = sizeof(SynergyDriverPdRelMove);
        break;
      }

      req = (SynergyDriverPdRelMove*)(devReq->memptr);

      inf->evpd.stat.read  = 0;
      inf->evpd.stat.abs   = 1;
      inf->evpd.stat.vst   = 0;
      inf->evpd.stat.inv   = 0;
      inf->evpd.stat.nodsp = 0;
      inf->evpd.xpos += req->x;
      if (inf->evpd.xpos < 0) {
        inf->evpd.xpos = 0;
      }
      else if (inf->evpd.xpos >= PDIN_XMAX) {
        inf->evpd.xpos = PDIN_XMAX - 1;
      }
      inf->evpd.ypos += req->y;
      if (inf->evpd.ypos < 0) {
        inf->evpd.ypos = 0;
      }
      else if (inf->evpd.ypos >= PDIN_YMAX) {
        inf->evpd.ypos = PDIN_YMAX - 1;
      }

      /*
       * ܥɥ饤ФǤϡԤʤΤǡKB/PDɥ饤Ф
       * ɬ⤤ͥ٥롼פ˽°ʤФʤʤ
       */
      snd_msg(inf->kpinput, (T_MSG*)&(inf->evpd));

      err = sizeof(SynergyDriverPdRelMove);
    }
    break;
  case DN_SYNERGY_PDWHEEL:
    {
      SynergyDriverPdWheel req;

      err = CheckDataCnt(devReq, sizeof(SynergyDriverPdWheel));
      if (err < 0) {
        goto err_ret;
      }
      else if (err == 0) {
	err = sizeof(SynergyDriverPdWheel);
        break;
      }

      req = *(SynergyDriverPdWheel*)(devReq->memptr);

      inf->evpd2.stat.read = 0;
      inf->evpd2.wheel     = req;

      /*
       * ܥɥ饤ФǤϡԤʤΤǡKB/PDɥ饤Ф
       * ɬ⤤ͥ٥롼פ˽°ʤФʤʤ
       */
      snd_msg(inf->kpinput, (T_MSG*)&(inf->evpd2));

      err = sizeof(SynergyDriverPdWheel);
    }
    break;
  case DN_SYNERGY_PDINPUT:
    {
      SynergyDriverPdInput* req = NULL;

      err = CheckDataCnt(devReq, sizeof(SynergyDriverPdInput));
      if (err < 0) {
        goto err_ret;
      }
      else if (err == 0) {
	err = sizeof(SynergyDriverPdInput);
        break;
      }

      req = (SynergyDriverPdInput*)(devReq->memptr);

      inf->evpd.stat.read   = 0;
      inf->evpd.stat.abs    = 1;
      inf->evpd.stat.vst    = 0;
      inf->evpd.stat.inv    = 0;
      inf->evpd.stat.nodsp  = 0;
      switch (req->btn) {
      case 1:
        inf->evpd.stat.main   = (req->on ? 1 : 0);
        inf->evpd.stat.qpress = 0;
        inf->evpd.stat.onebut = 0;
        break;
      case 2:
        inf->evpd.stat.main   = (req->on ? 1 : 0);
        inf->evpd.stat.qpress = (req->on ? 1 : 0);
        inf->evpd.stat.onebut = (req->on ? 1 : 0);
        break;
      case 3:
        inf->evpd.stat.sub    = (req->on ? 1 : 0);
        break;
      default:
        break;
      }

      /*
       * ܥɥ饤ФǤϡԤʤΤǡKB/PDɥ饤Ф
       * ɬ⤤ͥ٥롼פ˽°ʤФʤʤ
       */
      snd_msg(inf->kpinput, (T_MSG*)&(inf->evpd));

      err = sizeof(SynergyDriverPdInput);
    }
    break;
  case DN_SYNERGY_PDLEAVE:
    {
      SynergyDriverPdLeave* req = NULL;

      err = CheckDataCnt(devReq, sizeof(SynergyDriverPdLeave));
      if (err < 0) {
        goto err_ret;
      }
      else if (err == 0) {
	return sizeof(SynergyDriverPdLeave);
      }

      req = (SynergyDriverPdLeave*)(devReq->memptr);

      inf->evpd.stat.read   = 0;
      inf->evpd.stat.abs    = 1;
      inf->evpd.stat.vst    = 0;
      inf->evpd.stat.inv    = 1;
      inf->evpd.stat.nodsp  = 1;

      /*
       * ܥɥ饤ФǤϡԤʤΤǡKB/PDɥ饤Ф
       * ɬ⤤ͥ٥롼פ˽°ʤФʤʤ
       */
      snd_msg(inf->kpinput, (T_MSG*)&(inf->evpd));

      err = sizeof(SynergyDriverPdLeave);
    }
    break;
  case DN_SYNERGY_KEYINPUT:
    {
      SynergyDriverKeyInput* req = NULL;

      err = CheckDataCnt(devReq, sizeof(SynergyDriverKeyInput));
      if (err < 0) {
        goto err_ret;
      }
      else if (err == 0) {
	err = sizeof(SynergyDriverKeyInput);
        break;
      }

      req = (SynergyDriverKeyInput*)(devReq->memptr);

      inf->evkey.stat.read  = 0;
      inf->evkey.stat.press = (req->on ? 1 : 0);
      inf->evkey.keytop     = req->keytop;

      /*
       * ܥɥ饤ФǤϡԤʤΤǡKB/PDɥ饤Ф
       * ɬ⤤ͥ٥롼פ˽°ʤФʤʤ
       */
      snd_msg(inf->kpinput, (T_MSG*)&(inf->evkey));

      err = sizeof(SynergyDriverKeyInput);
    }
    break;
  default:
    err = toERR(EC_PAR, ED_DATANO);
    break;
  }
  if (err < ER_OK) {
    goto err_ret;
  }

  devRsp->datacnt = (W)err;

  return ER_OK;

err_ret:
  DEBUG_PRINT(("writeData err = %d\n", err));
  return err;
}


/*
 * ץ׵ν
 */
LOCAL ERR openDevice(SynergyDriverInfo* inf, ID devid)
{
  inf->opened = True;

  return ER_OK;
}


/*
 * ׵ν
 */
LOCAL ERR closeDevice(SynergyDriverInfo *inf)
{
  /* openreq = 0 ʤΤǡ OK */
  inf->opened = False;

  return ER_OK;
}


/*
 * ܡ׵ν
 */
static ERR abortRequest(SynergyDriverInfo *inf, DevReq *devReq)
{
  return ER_OK;
}


/*
 * ڥ׵ν
 *	ڥɻ˥ܡȤ줿׵ ER_MINTR ȤΤ§
 *	ꥢ̿Ǥϡڥ˼줿ǡϼƤޤᡢ
 *	ER_MINTR ֤Ƥ⤢ޤ̣ʤդˡץꥱ󤬤
 *	ʾ֤ˤʤäƤޤȤͤ롣äơER_MINTR ֤
 *	̾ΥܡȽƱ顼ɤ֤Ȥˤ롣
 */
static ERR suspendRequest(SynergyDriverInfo* inf)
{
  /* ڥɾ֤Ǥ뤳Ȥ򼨤 */
  Lock(&inf->lock);

  inf->suspended = True;

  return ER_OK;
}


/*
 * ꥸ塼׵ν
 */
static ERR resumeRequest(SynergyDriverInfo* inf)
{
  /* ڥɴʳ̵뤹 */
  if (!inf->suspended) {
    return ER_OK;
  }

  inf->suspended = False;

  /* ڥɾ֤Ȥ򼨤 */
  Unlock(&inf->lock);

  return ER_OK;
}


/*
 * ǥХ׵μ¹
 *	devReq ׵̤ devRsp ֤
 */
static ERR execDeviceRequest(SynergyDriverInfo *inf, DevRsp *devRsp, 
                             DevReq *devReq)
{
  ERR	err;

  switch ( devReq->cmd.cmd ) {
  case DC_OPEN:
    err = openDevice(inf, devReq->devid);
    break;

  case DC_CLOSE:
  case DC_CLOSEALL:
    err = closeDevice(inf);
    break;

  case DC_READ:
    err = readData(inf, devRsp, devReq);
    break;

  case DC_WRITE:
    err = writeData(inf, devRsp, devReq);
    break;

  case DC_ABORT:
    err = abortRequest(inf, devReq);
    break;

  case DC_SUSPEND:
    err = suspendRequest(inf);
    break;

  case DC_RESUME:
    err = resumeRequest(inf);
    break;

  default:
    /* ޥɥ顼 */
    err = toERR(EC_PAR, ED_CMD);
    DEBUG_PRINT(("devReq command(%d) err\n", devReq->cmd.cmd));
  }

  return err;
}


/*
 * ǥХ׵
 *	̾׵ü׵ξѤ
 */
ERR synergyAcceptDeviceRequest(SynergyDriverInfo *inf, UW reqptn)
{
  DevReq devReq;
  DevRsp devRsp;
  UW acpptn;
  RNO rdvno;
  INT size;
  ERR err;

  /* ʬ齪λϤʤ */
  for (;;) {
    if (inf->suspended) {
      /* ڥϡDC_RESUME ʳü׵
         դߤ */
      acpptn = reqptn & (D_NORM_PTN|D_CALPTN(DC_RESUME));
    } else {
      acpptn = reqptn;
    }

    /* ׵դ */
    err = acp_por(&rdvno, (VP)&devReq, &size, inf->acceptPort, acpptn);
    if (err < E_OK) {
      DEBUG_PRINT(("acp_por er = %d\n", err));
      continue;
    }

    /* ѥåȤν */
    devRsp.devid   = devReq.devid;
    devRsp.cmd     = devReq.cmd;
    devRsp.datano  = devReq.datano;
    devRsp.datacnt = 0;

    if (size != sizeof(DevReq)) {
      err = toERR(EC_PAR, ED_CMD);
      DEBUG_PRINT(("invalid devReq size (%d)\n", size));
      goto err_skip;
    }

    if ((D_CALPTN(devReq.cmd.cmd) & D_NORM_PTN) != 0) {
      /* ܡоݤȤʤ¹Ǥ뤳ȤϿ */
      inf->devReq = &devReq;

      /*
       * acp_por äƤ餳ޤǤδ֤ˡü׵
       * ¹ԤƤϤʤΤᡢԤ
       * 褦ʥƥॳȯԤƤϤʤ
       */

      /* ü׵᤬׵ᤵƤˤϡͥ褹
         褦˰ö¹Ը롣*/
      rot_rdq(TPRI_RUN);

      /* ̾׵νǤ뤳Ȥ򼨤 */
      Lock(&inf->lock);
    }

    /* ¹ */
    err = execDeviceRequest(inf, &devRsp, &devReq);

    if ((D_CALPTN(devReq.cmd.cmd) & D_NORM_PTN) != 0) {
      /* ̾׵νλ򼨤 */
      Unlock(&inf->lock);

      /* ܡоݤȤʤλȤϿ */
      inf->devReq = NULL;
    }

  err_skip:
    devRsp.error.err = err;

    /*  */
    err = rpl_rdv(rdvno, (VP)&devRsp, sizeof(DevRsp));
    if ( err < E_OK ) DEBUG_PRINT(("rpl_rdv er = %d\n", err));
  }

  return ER_SYS; /* 뤳ȤϤʤϤ */
}
